/* Specifications.java

		Specifies dimensions and performance of tracker detectors.

 Created by  Mike Ronan  Jan '99   Pass tracker specifications to TrkCandidate
                                   in recalculating track parameters.
 Modified by
   Nick Sinev    Mar '99    - Added areNew() and update(LCDEvent) methods
   Mike Ronan    Feb '00    - Simplify object model

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;

/** Specifies dimensions and performance of the tracking systems. */
public interface Specifications
{
    /** Get type of tracking detector. */
    public String getType();

    /** Get detector type as integer  Large = 1, Small = 2*/
    public int getDetectorType();
   
    /** Get inner radius of tracking detector. */
    public double getInnerRadius();
    /** Get outer radius of tracking detector. */
    public double getOuterRadius();
    /** Get half-length of tracking detector. */
    public double getHalfLength();
    /** Get half-length of tracking detector at radius r. */
    public double getHalfLength(double r);

    /** Get number of layers in tracking detector. */
    public int getNLayers();

    /** Get r-phi measurement resolution of tracking detector. */
    public double getRPhiResolution();
    /** Get z measurement resolution of tracking detector. */
    public double getZResolution();

    /** Get number of sigma for tracking roads. */
    public int getNSigmaTracking();

    /** Get inner radius of Vertex Detector. */
    public double getVxdInnerRadius();
    /** Get outer radius of Vertex Detector. */
    public double getVxdOuterRadius();
    /** Get half-length of Vertex Detector. */
    public double getVxdHalfLength();

    /** Get inner radius of Tracker. */
    public double getTrkInnerRadius();
    /** Get outer radius of Tracker. */
    public double getTrkOuterRadius();
    /** Get half-length of Tracker. */
    public double getTrkHalfLength();

    /** Get magnetic field of tracking detector. */
    public double getBField();

    /** Get extrapolation radius for coil. */
    public double getCoilExtrap();

    /** Get inner radius of central EM calorimeter. */
    public double get_EM_InnerRadius();
    /** Get outer radius of central EM calorimeter. */
    public double get_EM_OuterRadius();
    /** Get inner end plane of endcap EM calorimeter. */
    public double get_EM_InnerEndPlane();
    /** Get outer end plane of endcap EM calorimeter. */
    public double get_EM_OuterEndPlane();

    /** Get inner radius of central HAD calorimeter. */
    public double get_HAD_InnerRadius();
    /** Get outer radius of central HAD calorimeter. */
    public double get_HAD_OuterRadius();
    /** Get inner end plane of endcap HAD calorimeter. */
    public double get_HAD_InnerEndPlane();
    /** Get outer end plane of endcap HAD calorimeter. */
    public double get_HAD_OuterEndPlane();

    /** Get extrapolation radius for no field region. */
    public double getNoFieldExtrap();

    /** Get extrapolation radius for central Muon chamber. */
    public double getMuonExtrap();
    /** Get extrapolation plane for endcap Muon chambers. */
    public double getMuonEndExtrap();

    public double getLayerRadius(int layer);
    //public double getLayerOuterRadius(int layer);
    public double getLayerZ(int layer);

    /** Set detector r-phi resolution. */
    public void setRPhiResolution(double resolution);
    /** Set detector z resolution. */
    public void setZResolution(double resolution);

    /** Determine if specs have changed since last event. */
    public boolean areNew();

    /** Update specs if geometry has changed. */
    public void update(LCDEvent e);
}
