/* Tracker.java

		Find and reconstruct tracks in main tracker.

 Created by  Mike Ronan  Dec '98   Pick up LCD events within JAS framework and
                                   pass 3D points to pattern recognition software.
 Modified by
   Mike Ronan  Mar '00    - Add methods for controlling histogram packages.

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;
import hep.lcd.util.driver.*;


/**   Picks up LCD events within JAS framework and passes 3D points to pattern recognition software */
public interface Tracker extends Processor
{

    /** Get tracker specifications. */
    public Specifications getSpecifications();

    /** Get type of track hits. */
    public String getType();

    /** Setup Tracker parameters. */
    public void setupParameters();

    /** Set Vertex Detector. */
    public void setVertexDetector(VertexDetector detector);
    /** Set track fitter. */
    public void setFitter(TrackFitter fitter);

    /** Get Track finder. */
    public TrackFinder getFinder();
    /** Get Track finder */
    public TrackFinder getTrackFinder();

    /** Get number of track found. */
    public int getNTracks();
    /** Get track parameters. */
    public double[] getParameters(int itrk);
    /** Get the distance of closest approach (DOCA) of the track to the interaction point. */
    public double get_d0(int itrk);
    /** Get the phi azimuth of the point of closest approach. */
    public double get_phi0(int itrk);
    /** Get the reconstructed curvature of the track. */
    public double get_curv(int itrk);
    /** Get the distance in z from the point of closest approach to the interaction point. */
    public double get_z0(int itrk);
    /** Get the tangent of the dip angle of the track. */
    public double get_tandip(int itrk);

    /** Get number of points on a track. */
    public int getNPoints(int itrk);
    /** Get number of vertex detector points on a track. */
    public int getNVXDPoints(int itrk);
    /** Get number of tracker points on a track. */
    public int getNTrkPoints(int itrk);


    /** Get all the points on the track. */
    public int[] getTrackPoints(int itrk);

    /** Get a point on the track. */
    public double[] getTrackPoint(int itrk, int ipnt);

    /** Get the tracker hit for particular point on a track. */
    public TrackerHit getTrackHit(int itrk, int ipnt);

    public int getTrackHitID(int itrk,int ipnt);


    /** Returns pool of x-cordinates of hits. */
    public double[] getX();
    /** Returns pool of y-cordinates of hits. */
    public double[] getY();
    /** Returns pool of z-cordinates of hits. */
    public double[] getZ();
    /** Returns an array of markers of used hits. */
    public boolean[] getFlag();
    /** Returns array of starting point index for each layer. */ 
    public int[] getIpoi();
    /** Returns number of points in each layer. */
    public int[] getIlon(); 
    /** Returns tracker hits. */
    public TrackerHit[] getHits();


    /** Get Monte Carlo particle parent. */
    public MCParticle getMCParticle(int itrk);
    /** Get fraction of hits from parent MC particle. */
    public double getMCPFraction(int itrk);
    /** Get fraction of Tracker hits from parent MC particle. */
    public double getMCPTrackerFraction(int itrk);

    /** Set Monte Carlo particle parent. */
    public void setMCParticle(int itrk, MCParticle mcp);
    /** Put fraction of hits from parent MC particle. */
    public void putMCPFraction(int itrk, double fraction);
    /** Put fraction of Tracker hits from parent MC particle. */
    public void putMCPTrackerFraction(int itrk, double fraction);
   

    /** Returns status of track fitting. */
    public boolean fitSuccess(int itrk);
    /** Get overall chi-squared for the selected track. */
    public double getChi2(int itrk);
    /** Get number of degrees of freedom for the selected track. */
    public double getNDF(int itrk);
    /** Get track error matrix. */
    public double[][] getErrorMatrix(int itrk);

    /** Update fit status after fitting tracks. */
    public void updateFitStatus(int itrk, TrackFitter fitter);
    /** Replaces track parameters with fitter results. */
    public void replaceTrackParameters(int it, double[] newpar);


    /** Assigns given points to the track. */
    public int assignPoints(int it, double[] x, double[] y, double[] z, boolean[] KFlag,
			    int nLayers, int[] ipoi, int[] ilon,
			    double Ddzax, double Dmasu, int MaxPts,
			    int ipnt[][],
			    double Sang[], double Chis[], double ChisN[]);
/*
    public int assignPoints(int itrk,
			    int nLayers,
			    double Ddzax, double Dmasu, int MaxPts,
			    int ipnt[][],
			    double Sang[], double Chis[], double ChisN[]);
*/

    /** Turns on Tracker production histograms. */
    public void setHprod();
    /** Turns on Tracker test histograms. */
    public void setHtest();
    /** Turns on Tracker development histograms. */
    public void setHdev();
}
