/* TrackingMaterials.java

  This class extracts geometrical and materials information from detector description file
  for the given trajectory. It returns geometrical path lengths from the track origin to
  the radiation units weighted center of the gravity of each layer in detector, amount of 
  materials track encounter in each layer (in radiation length units) and inner radius of every 
  layer. It also provides the array of boolean variables which indicates for each layer 
  if it was reached by the track. There is also option to group layers in "superlayers" 
  (which is usefull for TPC with very large number of individual layers).   

  Author Nick Sinev (650)-926-2913    June, 1999

 Modified by
   Mike Ronan    May '01    - 

 */

package hep.lcd.recon.ztracking;

import hep.lcd.event.*;
//import hep.lcd.event.Track;
import hep.lcd.geometry.*;
import hep.lcd.geometry.component.*;
import hep.lcd.util.material.*;

import java.text.*;
import java.util.*;


public class TrackingMaterials
{
    boolean debug = false;
    int nCCDLrs = 5;

    final int MAXLAYERS = 160;
    Detector d;
    Track tr;
    TrackParameters tpar = new TrackParameters(0.,0.,0.,0.,0.);
    HelicalSwimmer swmr;
    double[] r0 = new double[3];
    double[] p0 = new double[3];
    int tchg;
    double[] pathlen = new double[MAXLAYERS];
    double[] thickns = new double[MAXLAYERS];
    double[] pathco = new double[MAXLAYERS];  
    double[] lradlen = new double[5];
    double[] ldedx = new double[5];
    double[] lthk = new double[5];
    double[] lradii=new double[MAXLAYERS];
    boolean[] active = new boolean[MAXLAYERS];

    int npnts=0;
    boolean useBP = false; 
    boolean useTPC = false;
    boolean useCCD = false;
    boolean largeDet = false;
    int ngrp = 1;
    int itpcr = 0;
    RadLengths rltab = new RadLengths();
    DeDx dedx = new DeDx();
    double halfpi = 0.5 * Math.PI;
    int strtv = 0;
    int strtt = 0; 

    public TrackingMaterials()
    {
	df.setMaximumFractionDigits(3);
    }

    public void setGroup(int ngr)
    {
	ngrp = ngr;
    }

    public double[] getLengths()
    {
	return pathlen;
    }

    public double[] getRadii()
    {
	return lradii;
    }

    public double[] getThick()
    {
	return thickns;
    }

    public double getThick(int ind)
    {
	return thickns[ind];  
    }

    public double getLayerThick(int ind)
    {
	int li = 0;
	if (ind > strtv-1) li = 1;
	if (ind > strtt-1) li = 2;
	return lradlen[li];
    }

    public double getDeDx(int ind)
    {
	int li = 0;
	if (ind > strtv-1) li = 1;
	if (ind > strtt-1) li = 2;
	return pathco[ind] * ldedx[li];
    }

    public double getLayerDeDx(int ind)
    {
	int li = 0;
	if (ind > strtv-1) li = 1;
	if (ind > strtt-1) li = 2;
	return ldedx[li];
    }


    public boolean[] getActive()
    {
	return active;
    }

    public int getNLayers()
    {
	return npnts;
    }

    public void putTrack(Track track)
    {
	d = Detector.instance();
	Field_Strength field = (Field_Strength) d.get("Field_Strength");
	double BField = field.getField();
	this.tr = track;
	for (int i=0; i<3; i++) {
	    p0[i] = tr.getMomentum()[i];
	    r0[i] = tr.getOrigin()[i];
	}
	tchg = tr.getCharge();
	double pt = Math.sqrt(p0[0]*p0[0] + p0[1]*p0[1]);
	double ph0 = Math.atan2(p0[1],p0[0]);
	tpar.put_phi0(ph0);
	double phor = ph0 + tchg * halfpi;
	if (Math.abs(Math.cos(phor)) > Math.abs(Math.sin(phor)))
	    tpar.put_d0(r0[0]/Math.cos(phor));
	else
	    tpar.put_d0(r0[1]/Math.sin(phor));
	tpar.put_omega(tchg * BField*0.003/pt);
	tpar.put_z0(r0[2]);
	tpar.put_s(p0[2]/pt);
	doTrack();
    }

    public void putParameters(TrackParameters tp)
    {
	d = Detector.instance();
	this.tpar = tp;
	doTrack();
    }

    private void doTrack()
    {
	for (int i=0; i<MAXLAYERS; i++) active[i] = false;
	df.setMaximumFractionDigits(5);
	double ph0 = tpar.get_phi0();
	double d0 = tpar.get_d0();
	Field_Strength field = (Field_Strength) d.get("Field_Strength");
	double BField = field.getField();
	if (debug) System.out.println("Field = "+BField);
	double pt = BField*0.3*.01/Math.abs(tpar.omega);
	tchg = tpar.omega>0. ? +1 : -1;
	p0[0]=pt*Math.cos(ph0);
	p0[1]=pt*Math.sin(ph0);
	p0[2]=pt*tpar.get_tandip();
	r0[0]=d0*Math.cos(ph0 + tchg * halfpi);
	r0[1]=d0*Math.sin(ph0 + tchg * halfpi);  
	r0[2] = tpar.get_z0(); 
	npnts = 0;
	double r_cyl = 100.;
	double z_pl = 100.;
	double bpr = 2.0;
	swmr = new HelicalSwimmer(BField,r_cyl,z_pl);
	swmr.swim(p0,r0,tchg);
	if (debug) System.out.println("extracting geometry");
	PropertySet bp = (PropertySet) d.get("Pipe_BARREL");
	useBP = (bp != null);
	PropertySet vtx = (PropertySet) d.get("MultiLayer_VERTX_Barrel");
	useCCD = (vtx != null);
	PropertySet tpc = (PropertySet) d.get("MultiLayer_TRACK_Barrel");
	useTPC = (tpc != null);
	if (useTPC) {
	    double innr = tpc.getDouble("inner_radius1");
	    largeDet=(innr > 15.);
	}
	calculateRadLen();
	int lind = 0;
	if (useBP) {
	    double bpri = bp.getDouble("inner_radius");
	    double bprit = bpri +0.1;
	    swmr.swim(bpri);
	    double lenin = swmr.getLength();
	    swmr.swim(bprit);
	    double lenout = swmr.getLength();
	    pathco[npnts] = 10. * (lenout-lenin);
	    bpr = bpri+lthk[lind];
	    lradii[npnts] = bpr;
	    swmr.swim(bpr);
	    pathlen[npnts] = swmr.getLength();   
	    active[npnts] = (swmr.getPlane() == swmr.PLANE_R);
	    npnts++;
	    lind++;
	}
	if (debug) System.out.println("BP");
	itpcr = 0;    
	if (useCCD) {
	    double nob = vtx.getDouble("num_of_barrels");
	    int inb = (int) nob;
	    int nl = 0;
	    for (int i = 1; i<inb+1; i++) {
		double ri = vtx.getDouble("inner_radius"+i);
		double rit = ri +0.1;
		swmr.swim(ri);
		double lenin = swmr.getLength();
		swmr.swim(rit);
		double lenout = swmr.getLength();
		pathco[npnts] = 10. * (lenout-lenin);
		double innr = ri + lthk[lind];
		z_pl = vtx.getDouble("z_outer"+i);
		lradii[npnts] = innr;
		double nlpb = vtx.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		swmr.swim(innr);
		pathlen[npnts] = swmr.getLength();
		active[npnts] = (swmr.getPlane() == swmr.PLANE_R);
		npnts++;
		nl+=inlpb;
	    }
	    lind++;
	    itpcr = nl;
	}
	if (useTPC) {
	    double nob = tpc.getDouble("num_of_barrels");
	    int inb = (int) nob;
	    int nl = 0;
	    int nlt = 0;
	    double ri = 0.;
	    double rout = 0.;
	    for (int i = 1; i<inb+1; i++) {
		double nlpb = tpc.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		nlt+= inlpb;
	    }

	    int nlin = 0;
	    int nla = 0;
	    while (nlin < nlt) {
		for (int i = 1; i<inb+1; i++) {
		    double innr = tpc.getDouble("inner_radius"+i)+lthk[lind];
		    double outr = tpc.getDouble("outer_radius"+i)+lthk[lind];
		    z_pl = tpc.getDouble("z_outer"+i);
		    double nlpb = tpc.getDouble("layers"+i);
		    int inlpb = (int) nlpb;
		    double lath = 0.;
		    if (inlpb > 0) lath = (outr-innr)/nlpb;
		    if (debug) System.out.println("nlpb= "+inlpb+" for i= "+i); 
		    for (int j = 0; j<inlpb; j++) {
			if (nlin%ngrp == 0) {
			    nla=0;
			    r_cyl = 0.;
			}
			r_cyl+= innr + j * lath;
			if(nlin%ngrp == 0) ri = innr+j*lath;
			if(nlin%ngrp == ngrp-1) rout = innr + (j+1) * lath;       
			nla++;
			if (nlin%ngrp == ngrp-1) {
			    swmr.swim(ri);
			    double lenin = swmr.getLength();
			    swmr.swim(rout);
			    double lenout = swmr.getLength();
			    active[npnts] = (swmr.getPlane() == swmr.PLANE_R);
			    pathco[npnts] = (lenout-lenin)/(rout-ri);
			    r_cyl/=nla;
			    swmr.swim(r_cyl,z_pl);
			    pathlen[npnts] = swmr.getLength();
			    lradii[npnts] = r_cyl;
			    npnts++;
			}
			nl++;
			nlin++;
		    }
		}
	    }
	}

	for (int i=0; i<npnts; i++) {
	    int li = 0;
	    if (i > strtv-1) li = 1;
	    if (i > strtt-1) li = 2;
	    thickns[i] = lradlen[li]*pathco[i];
	}
    }

    private void calculateRadLen()
    {
	int pnts = 0;
	if (useBP) {
	    strtv++;
	    strtt++;
	    MaterialList bpml = (MaterialList) d.get("Pipe_Material");
	    double brl = 0.;
	    lthk[pnts] = 0.;
	    ldedx[pnts] = 0.;
	    double thgr = 0.;
	    double thkp = 0.;
	    for (int i=0; i<bpml.getNMaterials(); i++) {
		Material bmat = bpml.getMaterial(i);
		String mat =  bmat.getName(); 
		double thk = bmat.getThickness();
		thgr+=0.5*thkp+0.5*thk;
		thkp = thk;
		double rt = thk/rltab.getRL(mat);
		double el = thk * dedx.getDeDx(mat);
		ldedx[pnts]+=el;
		lthk[pnts]+=thgr*rt;
		brl+=rt;
	    }
	    lthk[pnts]/=brl; 
	    lradlen[pnts] = brl;
	    if (debug)
		System.out.println("Pipe: lthk="+df.format(lthk[pnts])+" rad.thk="+df.format(brl));
	    pnts++;
	}
	if (useCCD) {
	    strtt+= nCCDLrs;
	    MaterialList vml = (MaterialList) d.get("VERTX_Barrel_Material");
	    double brl = 0.;
	    lthk[pnts]=0.;
	    ldedx[pnts]=0.;
	    double thgr = 0.;
	    double thkp = 0.;
	    for (int i=0; i<vml.getNMaterials(); i++) {
		Material vmat = vml.getMaterial(i);
		String mat =  vmat.getName();
		double thk = vmat.getThickness();
		double el = thk * dedx.getDeDx(mat);
		ldedx[pnts]+=el;
		thgr+=0.5*thkp+0.5*thk;
		thkp = thk;
		double rt = thk/rltab.getRL(mat);
		lthk[pnts]+=thk*rt;
		brl+=rt;
	    }
	    lthk[pnts]/=brl;
	    lradlen[pnts] = brl;
	    if (debug)
		System.out.println("VXD: lthk="+df.format(lthk[pnts])+" rad.thk="+df.format(brl));
	    pnts++;
	}
	if (useTPC) {
	    MaterialList tpml = (MaterialList) d.get("TRACK_Barrel_Material");
	    double brl = 0.;
	    lthk[pnts] = 0.;
	    ldedx[pnts]=0.;
	    double thgr = 0.; 
	    double thkp = 0.;
	    for (int i=0; i<tpml.getNMaterials(); i++) {
		Material tpmat = tpml.getMaterial(i);
		String mat = tpmat.getName();
		double thk = tpmat.getThickness(); 
		double el = thk * dedx.getDeDx(mat);
		ldedx[pnts]+=el;
		thgr+=0.5*thkp+0.5*thk;
		thkp = thk;
		double rt = thk/rltab.getRL(mat);
		lthk[pnts]+=thgr*rt;
		brl+=rt;
	    }
	    lthk[pnts]/=brl;
	    lradlen[pnts] = brl*ngrp;
	    lthk[pnts]*=ngrp;
	    pnts++;
	}
    }

    String ClassName = "TrackingMaterials::";

    DecimalFormat df = new DecimalFormat();
}
