/* SLDTrackFitter.java

   SLD Weight-Matrix (WM) fitter


  public void fit(int it,VertexDetector v,Tracker t)
    fit(it);
  public void fit(int it)
    countPoints();                         - Counts potential measurement points for the given track.  
    getTrkInters(parameters);              - 
    getTrackHits(it);                      - Extracts hits
      fillHits(layerhits,nlhits);          - 
   Fit tracks
    doIterate();                           - 
      getDerivatives();
      getSolution();
      getTrkInters(parameters);
      getResiduals(residuals);
      calculateChi2();


 Created by  Nick Sinev  May '99   
 Modified by
   Mike Ronan    May '01      - 

 */

package hep.lcd.recon.ztracking.fitters.sld;

import hep.lcd.event.*;
import hep.lcd.geometry.*;
import hep.lcd.geometry.component.Field_Strength;
import hep.lcd.recon.ztracking.*;

import hep.analysis.*;    // Histogram,...

import Jama.*;

//import java.lang.Double;
import java.text.*;
import java.util.*;


public class SLDTrackFitter extends AbstractTrackFitter
{
    final int MaxTrackHits  = 200;
    final int MaxLayers = 300;
    //final int MaxLayerHits = 1000;  // <== should be only 2 or 3 ???
    final int MaxLayerHits = 3;
    // final int MaxLayerHits = 5;

    int NVertexLayers = 5;
    int NTrackerLayers = 144;

    int nvhits;
    int nthits;

    double GoodChi2 = 10.;
    double[] r0 = new double[3];
    double[] p0 = new double[3];

    int NIterations = 2;  // Number of iterations

    Matrix wm;
    Matrix wp;
    double[][] wma;
    boolean acceptAll = false;

    Detector det;
    double BField = 0.;

    double[] dpar=new double[5]; 
    double[] Ym = new double[5];
    double[][] WPmn = new double[5][5];

    TrackerHit[] thits;


    public SLDTrackFitter()
    {
	this(false);
    }
    /** */
    public SLDTrackFitter(boolean hist)
    {
	super(hist);

	System.out.println(ClassName+"  version 1.1.6.2");

	Fitter = "SLD";

	df.setMaximumFractionDigits(4);
    }

    /** Invoked in setting up TrackFitter to set fitting strategy. */
    public void setupParameters()
    {
	if (debug) System.out.println(ClassName+"setupParameters:  ");

	useVXD = (vxd != null);
	useTracker = (tracker != null);
	if (!useVXD && !useTracker) {
	    System.err.println(ClassName+" can't fit without a tracker device!");
	    return;
	} 
    }


    int itrk;

    public void fit(int it, VertexDetector v, Tracker t)
    {
	itrk = it;
	this.vxd = v;
	this.tracker = t;
	if (first) {
	  if (!useVXD) System.out.println(" Not using Vertex detector!");
	  if (!useTracker) System.out.println(" Not using Tracker!");
	}

	fit(it);

	first = false;
    }

    int nEvt = 0;

    public void fit(int it)
    {
	success = false;  
	badTrack = false;

	// if (it==0||it==35) debug = true;
	// else debug = false;
	if (it==0) nEvt++;
	//if (nEvt==78 && it==19) debug = true;
	//else debug = false;
       	//if (nEvt==268) debug = true;

	// Set detector and get magnetic field
	if (it==0) {
	  det = Detector.instance();
	  Field_Strength field = (Field_Strength) det.get("Field_Strength");
	  BField = field.getField();
	}

	if (debug) System.out.println("Fitting track # "+it);
	if (Hist) {
	  if (it==0) HEvents++;
	  HistogramFolder.setDefaultFolder(defaultFolder);
	}

	// Get track parameters from Tracker
	double[] trkpar = tracker.getParameters(it);

	// Internal Track parameters
	parameters = new TrackParameters(trkpar);

	// Enter track parameters into Weight matrix
	putTrackParameters(parameters);

	// Require that track passes Weight matrix checks
	if (check) {

	  double r_cyl = 100.;
	  double z_pl = 500.;
	  swmr.setBField(BField);
	  swmr.swim(parameters,r_cyl,z_pl);         

	  if (debug) System.out.println(" Count fit points");
	  countPoints();
	  if (Hstd && HEvents<=MaxTrkHistEvts) {
	    Histogram.find("NTot fit points/trk").fill(nlrs);
	    Histogram.find("NVXD fit points/trk").fill(nvlrs);
	    Histogram.find("NTrk fit points/trk").fill(ntlrs);
	  }

	  if (debug) System.out.println(" Calculating intersections (trkz[],trkphi[] for layer radii lrsr[]) ");
	  getTrkInters(parameters);

	  if (Hpnts && HEvents==1 && itrk==0) {
	    for (int i=1; i<=nvlrs; i++) {
	      Histogram.find("Vtx r vs z").fill(trkz[i],lrsr[i]);
	      Histogram.find("Vtx r vs phi").fill(trkphi[i],lrsr[i]);
	      Histogram.find("Trk r vs z").fill(trkz[i],lrsr[i]);
	      Histogram.find("Trk r vs phi").fill(trkphi[i],lrsr[i]);
	    }
	    for (int i=1+nvlrs; i<=nlrs; i++) {
	      Histogram.find("Trk r vs z").fill(trkz[i],lrsr[i]);
	      Histogram.find("Trk r vs phi").fill(trkphi[i],lrsr[i]);
	    }
	  }

	  if (debug) System.out.println(" Extracting Track hits"); 
	  getTrackHits(it);

	  if (!badTrack) {
	    if (debug) System.out.println(" Starting iterations");
	    doIterate();
	  }
	}
	if (Hist) HistogramFolder.setDefaultFolder("..");
    }

    boolean check;

    public void putTrackParameters(TrackParameters parameters)
    {
	if (debug) System.out.println(ClassName+"putTrackParameters: Parameters = ("+parameters+")");
	track = new ReconstructedTrack(BField,parameters);
	check = Wm.putTrack(track,debug);
	if (!check) return;
	initArrays();
    }

    void initArrays()
    {
	wm = Wm.getMatrix();
	wma = wm.getArray();
	//if (debug) System.out.println(ClassName+" Weight matrix calculated.");
    }


    void countPoints()
    /* ---------------------------------------------------------------------------------------------- 
       |   This routine count potential measurement points for the given track. Currently it considers
       |   point as potential measurement point only if it is within barrel tracker. 
       |   The entire fitter so far can handle only cylindrical detector. In the future endcaps will
       |   be incorporated.
       |   Routine takes care about grouping layers for large detector, so that about 12 measurement
       |   points will be generated for large tracker.
       |   It also gives estimate for layers active radii. This radii will be corrected during hit
       |   list filling, when real hitt coordinate will be used. For now we will be using geometry file
       |   inner layer radius + 1/2 of layer thickness as an estimation for hit radius.
       |_______________________________________________________________________________________________
    */  
    {
	//boolean ldebug = false;
	int wml = wma[0].length;

	nvlrs = 0;
	PropertySet vtx = (PropertySet) det.get("MultiLayer_VERTX_Barrel");
	if (vtx != null) {
	    //if (ldebug) System.out.println("  Processing VXD"); 
	    double nob = vtx.getDouble("num_of_barrels");
	    int inb = (int) nob;
	    for (int i = 1; i<=inb; i++) {
		double innr = vtx.getDouble("inner_radius"+i);
		double outr = vtx.getDouble("outer_radius"+i);
		double nlpb = vtx.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		double dr = (outr-innr)/nlpb;
		//if (ldebug) System.out.println("   barrel "+i+" layers "+inlpb+" in.r "+innr+" out.r "+outr);
		for (int j=0; j<inlpb; j++) {
		    double lr = innr + j * dr;
		    lrsr[1+nvlrs] = slrsr[1+nvlrs] = lr;  // Use all Vertex detector measurement points
		    nvlrs++;
		}
	    }
	}
	if (Hdet) Histogram.find("NVtx Layers").fill(nvlrs);

	ntlrs = 0;
	PropertySet tracker = (PropertySet) det.get("MultiLayer_TRACK_Barrel");
	if (tracker != null) {
	    //if (ldebug) System.out.println("  Processing TRACKER");
	    double nob = tracker.getDouble("num_of_barrels");  
	    int inb = (int) nob;

	    int nlt = 0;
	    for (int i = 1; i<=inb; i++) {
		double nlpb = tracker.getDouble("layers"+i);
		int inlpb = (int) nlpb;
		nlt+=inlpb;
	    }

	    npgrp = 1;
	    if (nlt > 36) npgrp = nlt/ngrps;
	    //if (ldebug) { text = "   Number of barrels: "+inb+"\n     Total number of layers: "+nlt+"\n     Number of layers in each TRACKER group is "+npgrp; System.out.println(text); }

	    int nlin = 0;
	    int nla = 0;
	    double rl = 0.;
	    boolean okswim = true;
	    while ((nlin < nlt) && okswim) {
		for (int i=1; i<=inb; i++) {
		    double innr = tracker.getDouble("inner_radius"+i);
		    double outr = tracker.getDouble("outer_radius"+i);
		    double z_pl = tracker.getDouble("z_outer"+i);
		    double lthk = 0.;
		    double nlpb = tracker.getDouble("layers"+i);
		    int inlpb = (int) nlpb;

		    if (inlpb > 0) lthk = (outr-innr)/nlpb;
		    for (int j=0; j<inlpb; j++) {
			slrsr[1+nvlrs+nlin] = innr + j * lthk + 0.5 * lthk;  // Layer radius
			if (nlin%npgrp == 0) {
			    nla = 0; rl = 0.;
			}
			nla++;
			rl+= slrsr[1+nvlrs+nlin];
/*
			if (nlin%npgrp == 0) {
			    rl/=nla;
			    lrsr[1+nvlrs+ntlrs] = rl;  // Average measurement point radius
			    //if (debug && okswim) System.out.println("   Swimming track out to r="+df.format(rl));
			    swmr.swim(rl,z_pl);
			    if (swmr.getPlane()==swmr.PLANE_R) ntlrs++;
			    else okswim = false;
			    if (debug && okswim)
			      System.out.println("   Swam track out to layer "+nlin+" at r="+df.format(slrsr[1+nvlrs+nlin]));
			}
*/
/**/
			if (nlin%npgrp == npgrp-1) {
			    rl/=nla;
			    lrsr[1+nvlrs+ntlrs] = rl;  // Average measurement point radius
			    //if (debug && okswim) System.out.println("   Swimming track out to r="+df.format(rl));
			    swmr.swim(rl,z_pl);
			    if (swmr.getPlane()==swmr.PLANE_R) ntlrs++;
			    else okswim = false;
			    if (debug && okswim)
			      System.out.println("   Swam track out to layer "+nlin+" at r="+df.format(slrsr[1+nvlrs+nlin]));
			}

			nlin++;
			if ((1+nvlrs+ntlrs)*2 >= wml) okswim=false; 
			if (nlin > nlt) okswim = false; 
			if (!okswim) break;      
		    }
		    if (!okswim) break;
		}
		if (!okswim) break;
	    }
	}
	if (Hdet && HEvents<=MaxTrkHistEvts) {
	    Histogram.find("NTrk Super Layers").fill(ntlrs);
	    Histogram.find("NTrk Groups").fill(npgrp);
	}

	//if (debug) System.out.println(ClassName+"countPoints: nvlrs = "+nvlrs+" ntlrs = "+ntlrs);
	if (debug) System.out.println("  nvlrs = "+nvlrs+" ntlrs = "+ntlrs);
	nlrs = nvlrs+ntlrs; 
	if (nlrs < 5) {
	    badTrack = true;
	    if (debug) System.out.println(" Track too short, only " +nlrs+" possible points available for fitting");
	}
    }     

    /** Calculates intersections */
    void getTrkInters(TrackParameters tp)
    {
	double[] fr0 = new double[3];
	double[] fp0 = new double[3];

	double fpt = BField * 0.003 /Math.abs(tp.omega);
	fr0[0] = tp.d0 * Math.cos(tp.phi0+halfpi);
	fr0[1] = tp.d0 * Math.sin(tp.phi0+halfpi);
	fr0[2] = tp.z0;
	fp0[0] = fpt * Math.cos(tp.phi0);
	fp0[1] = fpt * Math.sin(tp.phi0);
	fp0[2] = fpt * tp.s;
	int chrg = (tp.omega > 0) ? 1 : -1 ; 

	swmr.swim(fp0,fr0,chrg);

	for (int i=1; i<=nlrs; i++) {
	    double rl = lrsr[i];

	    swmr.swim(rl);
	    double[] xx = swmr.getIntersect();

	    trkz[i] = xx[2];
	    trkphi[i] = Math.atan2(xx[1],xx[0]);
	}
    }


    void getTrackHits(int it)
    {
	// Number of measurement points
	int nps = 1+nlrs;

	for (int i=0; i<nps; i++) active[i] = false;       
	for (int i=0; i<MaxLayers; i++) nlhits[i] = 0;

	double maxvr = slrsr[nvlrs]+0.2;                  // Note: HARD CODED
 
	int nvh = vxd.getNPoints(it);
	if (Htrks && HEvents<=MaxTrkHistEvts) Histogram.find("VXD nhits/trk").fill(nvh);

	for (int ip=1; ip<=nvh; ip++) {
	    double point[];
	    TrackerHit vh;
	    point = vxd.getTrackPoint(it,ip);
	    vh = vxd.getTrackHit(it,ip);
	    int nlr = vh.getLayer();
	    //if (Hhits && HEvents<=MaxHitHistEvts) Histogram.find("VXD layer").fill(nlr);
	    nlr+=1;
	    if (nlhits[nlr]<MaxLayerHits && nthits<MaxTrackHits) {
	      someHit hit = new someHit();
	      hit.putPoint(point);
	      layerhits[nlr][nlhits[nlr]]=hit;
	      nlhits[nlr]++;
	    }
	}
	if (Hhits && HEvents<=MaxHitHistEvts)
	    for (int n=1;n<=NVertexLayers;n++) Histogram.find("VXD nhits/layer").fill(nlhits[n]);

	int nth = tracker.getNTrkPoints(it);
	if (Htrks && HEvents<=MaxTrkHistEvts) Histogram.find("Trk nhits/trk").fill(nth);

	for (int ip=1; ip<=nth; ip++) {
	    double point[] = tracker.getTrackPoint(it,ip);
	    double r = Math.sqrt(point[0]*point[0] + point[1]*point[1]);
	    TrackerHit th = tracker.getTrackHit(it,ip);
	    int nlr = th.getLayer();
	    //if (Hhits && HEvents<=MaxHitHistEvts) Histogram.find("Trk layer").fill(nlr);
	    nlr+=1+NVertexLayers;

	    // Require radius of tracker hit be larger than vertex detector max. radius
	    if (r > maxvr && nlhits[nlr]<MaxLayerHits && nthits<MaxTrackHits) {
		someHit hit = new someHit();
		hit.putPoint(point);
		layerhits[nlr][nlhits[nlr]]=hit;
		nlhits[nlr]++;
	    }
	}
	if (Hhits && HEvents<=MaxHitHistEvts)
	    for (int n=1+NVertexLayers;n<=NVertexLayers+NTrackerLayers;n++) Histogram.find("Trk nhits/layer").fill(nlhits[n]);

	if (debug && printPoints) {
	    System.out.println("  Track # "+it);
	    for (int i=0; i<MaxLayers; i++) {
	      if (i<NPntsPrint) {
	      // if (i<NPntsPrint || i>=1+nvlrs+(ntlrs-1)*npgrp+1) {
		for (int n=0;n<nlhits[i];n++) {
		    someHit hit = layerhits[i][n];
		    double lrd = Math.sqrt(hit.pnt[0]*hit.pnt[0]+hit.pnt[1]*hit.pnt[1]);
		    swmr.swim(lrd);
		    double swpnt[] = swmr.getIntersect();
		    double r = Math.sqrt(swpnt[0]*swpnt[0]+swpnt[1]*swpnt[1]);
		    System.out.println("   layer "+i+" hit #"+n+" (r = "+df.format(lrd)+") x="+df.format(hit.pnt[0])+
				       " y="+df.format(hit.pnt[1])+" z="+df.format(hit.pnt[2])
				       +"\t track: x="+df.format(swpnt[0])+" y="+df.format(swpnt[1])+
				       " z="+df.format(swpnt[2])+"  (r = "+df.format(r)+")");
		}
	      }
	      if (i==NPntsPrint) System.out.println("   ...");
	    }
	}
       
	fillHits(layerhits,nlhits);
    } 

    void fillHits(someHit[][] layerhits,int[] nlhits)
    {
	if (debug) System.out.println("  Fill hits");

	double[] sltphi = new double[MaxLayers];
	double[] sltz = new double[MaxLayers];

	double[] bestch2 = new double[MaxLayers];      
	double[] bestmch= new double[MaxLayers];     
	for (int i=0; i<MaxLayers; i++) bestmch[i] = 99999999.;

	double[] fr0 = new double[3];
	double[] fp0 = new double[3];
	double fpt = BField * 0.003 /Math.abs(parameters.omega);

	// Get correct radii for active layers - take them from hits coordinates
	int nlayers = 1+nvlrs+ntlrs*npgrp; double rmax = 0.;
	for (int n=1; n<=nlayers; n++) {       
	    if (nlhits[n] > 0) {
		someHit th = layerhits[n][0];
		double[] hc=th.getPoint();
		double r = Math.sqrt(hc[0]*hc[0] + hc[1]*hc[1]);
		slrsr[n] = r;
		if (r>rmax) rmax = r;
		if (Hdet && HEvents<=MaxHitHistEvts) Histogram.find("Hits r layer").fill(r);
	    }
	}
	if (debug) System.out.println("   Layer radii extracted (rmax = "+df.format(rmax)+")");

	// now swim track through correct radii	
	fr0[0] = parameters.d0 * Math.cos(parameters.phi0+halfpi);
	fr0[1] = parameters.d0 * Math.sin(parameters.phi0+halfpi);
	fr0[2] = parameters.z0;
	fp0[0] = fpt * Math.cos(parameters.phi0);
	fp0[1] = fpt * Math.sin(parameters.phi0);
	fp0[2] = fpt * parameters.s;
	int chrg = (parameters.omega > 0) ? 1 : -1 ; 
	swmr.swim(fp0,fr0,chrg);

	for (int i=1; i<=nlayers; i++) {
	    double rl = slrsr[i];
	    swmr.swim(rl);
	    double[] xx = swmr.getIntersect();
	    sltz[i] = xx[2];
	    sltphi[i] = Math.atan2(xx[1],xx[0]);
	}


	// Find best point for each measurement group.
	if (debug) System.out.println("   Find best point for each measurement group.");

	int nps = 1+nlrs;     // # points
	int avln = 0;         // ?average? layer #
	int pin = 0;          // point #
	int pinp = pin;       // best point ?
	int wml = wma[0].length;

	double adevrp=0.,adevz=0.,aver=0.,nhing=0.;  // Group averaging: dev rphi,z,r # things??
	for (int n=1; n<=nlayers; n++) {
	    pin = n;
	    if (n>nvlrs) pin = 1+nvlrs+(n-nvlrs-1)/npgrp;
	    if ((pin+nps)>=wml) break;
	    //if (debug) System.out.println("     pin,nps = "+pin+","+nps);

	    int imin = 0;
	    double chmin = 0.;     
	    if (debug && n<NPntsPrint)
		System.out.print("     Layer #"+n+" has "+nlhits[n]+" hit(s)");
	    if (nlhits[n] > 0) {
		double[] Chi2 = new double[MaxLayerHits];
		for (int i=0; i<nlhits[n]; i++) {
		    someHit hit = layerhits[n][i];
		    double[] p = hit.getPoint();
		    double hphi = Math.atan2(p[1],p[0]);
		    double drphi = (hphi-sltphi[n])*slrsr[n];
		    double dz = p[2] - sltz[n];
		    //if (debug) System.out.print(" "+df.format(drphi)+","+df.format(dz)+" ");
		    Chi2[i] = wma[pin][pin] * drphi * drphi + wma[nps+pin][nps+pin] * dz * dz;
		}

		chmin = Chi2[0];
		imin = 0;
		for (int i=1; i<nlhits[n]; i++) {
		    if (Chi2[i] < chmin) {
			chmin = Chi2[i];
			imin = i;
		    }
		}
	    }
	    bestch2[n] = chmin;
	    if (debug && n<NPntsPrint)
		System.out.println(" with min. chi2 = "+df.format(chmin));

	    pin = (n < 1+nvlrs) ? n : 1+nvlrs+avln; 
	    if ((nlhits[n] > 0) && ((chmin < 100000.)||acceptAll) 
		&& ((n > nvlrs) || (useVXD))) {
		active[pin] = true;
	    }
	    if (pin != pinp) {
		bestmch[pin] = chmin;
	    }
	    if (bestmch[pin] > chmin) bestmch[pin] = chmin;
	    pinp = pin;
	    boolean lwh = ((nlhits[n] > 0) && ((chmin < 100000.)||acceptAll));       
	    double[] hc = {0.,0.,0.};
	    double hphi = 0.; 
	    if (lwh) {
		for(int i=0; i<3; i++)  
		    hc[i] = layerhits[n][imin].getPoint()[i];
		hphi = Math.atan2(hc[1],hc[0]);
	    }

	    hitsphi[n] = 0.;
	    hitsz[n] = 0.; 
	    //if (debug) System.out.println("       pin (n,avln) = "+pin+" ("+n+","+avln+")");
	    if ((n <= nvlrs) && lwh) {
		lrsr[n] = Math.sqrt(hc[0]*hc[0]+hc[1]*hc[1]);
		hitsphi[n] =  hphi;
		hitsz[n] = hc[2];
	    }
	    if (n > nvlrs) {
		if ((n-nvlrs-1)%npgrp == 0) {
		    adevrp = 0.;
		    adevz = 0.;
		    aver = 0.;
		    nhing = 0.;
		}
		if (lwh && ((bestch2[n] < 100000.)||acceptAll)) { 
		    adevrp+= (hphi-sltphi[n])*slrsr[n];
		    adevz+= hc[2] - sltz[n];
		    aver+=slrsr[n];
		    nhing+=1.;
		} 
		if ((n-nvlrs-1)%npgrp == (npgrp-1)) {
		    if (nhing > 0.) {
			adevrp/=nhing;
			adevz/=nhing;
			aver/=nhing;
		    } 
		    swmr.swim(aver);
		    double[] xx = swmr.getIntersect();
		    hitsz[pin] = xx[2] + adevz;
		    hitsphi[pin] = Math.atan2(xx[1],xx[0]) + adevrp/aver;
		    lrsr[pin] = aver;
		    avln++;
		}
	    }
	}
	if (debug) {
	    for (int i = 0; i<1+nlrs; i++) {
	      if (i==0) System.out.println("    Skip point #0 - No beam constaint");
	      else
		System.out.println("    Best chi2 for meas. pnt #"+i+" is "+df.format(bestmch[i]));
	    }
	}

	NDoF = 0.;
	for (int i = 1; i<=nlrs; i++) {
	    active[i] = (active[i] && Wm.getActive()[i]);  
	    if (active[i]) NDoF+=1.;
	}
	NDoF=2.* NDoF - 5.;
	if (debug) System.out.println("   NDoF = "+df.format(NDoF));

	// if((NDoF < 4.) || (NDoF < 0.6*(2*(nlrs) - 5)))
	if (NDoF < 3.) {
	    if (debug) System.out.println(" Too few points: NDoF = "+NDoF); 
	    badTrack = true;
	}
    }

    void getResiduals(double[] res)
    {
	int nps = 1+nlrs;
	for (int i=0; i<nps; i++) {
	    if (i==0) { res[i]=0; res[i+nps] = 0; }
	    if (!active[i]) { res[i]=0; res[i+nps] = 0; } 
	    if ((i != 0) && active[i]) {
		res[i] = (trkphi[i]-hitsphi[i]) * lrsr[i];
		res[i+nps] = trkz[i] - hitsz[i];
	    }
	}
	if (printResiduals) {
	    System.out.println("  Residuals: ");
	    //for (int i=7; i< nps; i++)
	    for (int i=0; i<nps; i++)
		System.out.println("   point "+i+" trkphi= "+df.format(trkphi[i])+
				   "  hitphi= "+df.format(hitsphi[i])+" res= "+df.format(res[i]));
	    printResiduals = false;
	}
    }

    void getDerivatives()
    {
	int nps = 1+nlrs;

	getTrkInters(parameters);
	getResiduals(residuals);

	double[] newres = new double[2*nps];
	double norm = 1./delpar;

	parameters.d0+=delpar;
	getTrkInters(parameters);
	getResiduals(newres);
	for (int i=0; i<2*nps; i++)
	    derivatives[i][0] = norm * (newres[i]-residuals[i]);
	parameters.d0-=delpar;

	parameters.phi0+=delpar;
	getTrkInters(parameters);
	getResiduals(newres);
	for (int i=0; i<2*nps; i++)
	    derivatives[i][1] = norm * (newres[i]-residuals[i]);
	parameters.phi0-=delpar;

	parameters.omega+=delpar;
	getTrkInters(parameters);
	getResiduals(newres);
	for(int i=0; i<2*nps; i++)
	    derivatives[i][2] = norm * (newres[i]-residuals[i]);
	parameters.omega-=delpar;

	parameters.z0+=delpar;
	getTrkInters(parameters);
	getResiduals(newres);
	for(int i=0; i<2*nps; i++)
	    derivatives[i][3] = norm * (newres[i]-residuals[i]);
	parameters.z0-=delpar;

	parameters.s+=delpar;
	getTrkInters(parameters);
	getResiduals(newres);
	for(int i=0; i<2*nps; i++)
	    derivatives[i][4] = norm * (newres[i]-residuals[i]);
	parameters.s-=delpar;
    }


    void getSolution()        
    {
	boolean ldebug = false;

	int nps = 2*(1+nlrs);
	int wml = wma[0].length;
	if (nps > wml) System.err.println(ClassName+" error: weight matrix length "+wml+" while nps="+nps);

	double ym = 0.;
	for (int m=0; m<5; m++) {
	    for (int n=0; n<5; n++) {
		double me = 0.;
		ym = 0.;
		for (int i=0; i<nps; i++) {
		    for (int j=0; j<nps; j++) {
			me+=derivatives[i][m]*wma[i][j]*derivatives[j][n];
			if (n==4) ym+= derivatives[i][m]*wma[i][j]*residuals[j];
		    }
		}
		WPmn[m][n] = me;
	    }
	    Ym[m] = -ym;
	}

	wp = new Matrix(WPmn);
	int nrs = 5;
	// wp.print(8,6);
	Matrix y = new Matrix(Ym,nrs);
	//  y.print(8,6);
	Matrix dp = wp.solve(y);

	if (ldebug) {
	  System.out.println(" Parameter increments: ");
	  dp.print(8,6);
	  // System.out.println("Error matrix: ");
	  //     erma.print(9,7);
	  if (success)
	    for (int i=0; i<5; i++)
		System.out.println("Error in parameter "+i+" is "+Math.sqrt(erma.get(i,i)));
	}

	for (int i=0; i<5; i++)
	    dpar[i] = dp.get(i,0);
    }

    void calculateChi2()
    {
	int nps = 2 * (1+nlrs);
	chi2 = 0.;               
	for (int i=0; i<nps; i++) {
	    int lin = 0;
	    if (i < nps/2) lin = i;
	    if (i > nps/2-1) lin = i - nps/2; 
	    if (i > (wma[0].length - 1)) break;
	    for (int j=0; j<nps; j++) {
		int ljn = 0;
		if (j<nps/2) ljn = j;
		if (j>nps/2-1) ljn = j-nps/2;
		if (j > (wma[0].length -1)) break;
		if (active[lin] && active[ljn]) 
		    chi2+= residuals[i]*wma[i][j]*residuals[j];
	    }
	}
	if ((chi2 > veryLarge-1.) ||(Double.toString(chi2).equals("NaN")))
	    chi2 = veryLarge;
    }

    void doIterate()
    {
	success = false;
	badTrack = false;

	for (int nit = 0; nit<NIterations; nit++) {
	    if (nit == 0) delpar = 0.0001;
	    else delpar = 0.00001;
	    getTrkInters(parameters);
	    getResiduals(residuals);
	    calculateChi2();
	    if (chi2 > veryLarge-1.) {
		if (debug) System.out.println("  very large Chi2 of candidate track!");
		break;
	    }

	    if (debug) System.out.println("  iteration #"+nit+" chi2= "+df.format(chi2));
	    getDerivatives();

	    getSolution();

	    for (int i=0; i<5; i++) {
		double maxdev = 0.1;
		if ((i==0) || (i==3)) maxdev = 1.0; 
		if (Math.abs(dpar[i]) > maxdev) {
		    badTrack = true;
		    if (debug) System.out.println(ClassName+" too large change in parameter "+i+" dpar="+df.format(dpar[i]));
		    chi2 = veryLarge;
		    return;
		} 
	    }
	    parameters.d0+=dpar[0];
	    parameters.phi0+=dpar[1];
	    parameters.omega+=dpar[2];
	    parameters.z0+=dpar[3];
	    parameters.s+=dpar[4];
	}

	getTrkInters(parameters);

	getResiduals(residuals);

	calculateChi2();

	if (NDoF > 1.) {
	    if (debug) System.out.println("  final chi2 = "+df.format(chi2)+" for NDoF = "+NDoF+" ("+df.format(chi2/NDoF)+")");

	    if (chi2/NDoF < GoodChi2) success = true;
	    else if (debug) System.out.println("   Too large  chi2/ndf: "+df.format(chi2/NDoF));

	    if (success) erma = wp.inverse();

	    if (debug && (chi2/NDoF > 10.)) {
		int nactv=0;
		int nactt = 0;
		for (int i=1; i<=nvlrs; i++)
		    if (active[i]) nactv++;
		for (int i=1+nvlrs; i <= nlrs; i++)
		    if (active[i]) nactt++;
		System.out.println("Bad fit ... NDoF= "+NDoF+" active vxd lrs "+nactv+
				   " active tracker lrs "+nactt);
	    }
	}
	if (Hpnts && HEvents<MaxPntHistEvts) {
	    int nps = 1+nlrs;
	    for (int i=1;i<nps;i++) {
		Histogram.find("Residuals xy").fill(residuals[i]);
		Histogram.find("Residuals z").fill(residuals[i+nps]);
	    }
	}
    }

    class someHit
    {
	public double[] pnt={0.,0.,0.};

	public someHit()
	{
	}

	public void putPoint(double[] pnt)
	{
	    this.pnt[0]=pnt[0];
	    this.pnt[1]=pnt[1];
	    this.pnt[2]=pnt[2];
	}

	public void putX(double x)
	{
	    this.pnt[0] = x;
	}

	public void putY(double y)
	{
	    this.pnt[1] = y;
	}

	public void putZ(double z)
	{
	    this.pnt[2] = z;
	}

	public double[] getPoint()
	{
	    return pnt;
	}

    } 


    String ClassName = "SLDTrackFitter::";

    // Inline debugging and analysis
    boolean debug = false;

    // Printing
    boolean printPoints = true;  int NPntsPrint = 12;  // Print out to inner most point of triplets
    boolean printResiduals = false;

    DecimalFormat df = new DecimalFormat();


    // Histogramming
    boolean Hdet = false;


    double halfpi = 0.5 * Math.PI;
    double delpar = 0.00001;

    // Tracking layers
    int nlrs,nvlrs,ntlrs,npgrp;
    final int ngrps = 12;
    double[] slrsr = new double[MaxLayers];     // Layer radii
    double[] lrsr = new double[20];             // Measurement point radii
    boolean[] active = new boolean[MaxLayers];  // Active layers

    // Track intersections
    double[] trkphi = new double[MaxLayers];
    double[] trkz = new double[MaxLayers];

    // Track hits
    int[] nlhits=new int[MaxLayers];
    someHit[][] layerhits = new someHit[MaxLayers][MaxLayerHits];

    double[] hitsphi = new double[MaxLayers];
    double[] hitsz = new double[MaxLayers];

    // Track fitting
    double[] residuals = new double[2*MaxLayers];
    double[][] derivatives = new double[2*MaxLayers][5]; 
    /* 
       note: number of residuals and derivatives equals to 2*(npoints+1), where npoints is number
       of points on track. Residuals and derivatives in rphi are the first, in z - next. The
       very first residual is always = 0., because first point corresponds to beam pipe, for which
       there is no measurement, but it is included in weight matrix calculations.
    */


    HelicalSwimmer swmr = new HelicalSwimmer();
    WeightMatrix Wm = new WeightMatrix(); 
}
